#include "stdexcept"
#include "stddef.h"
#include <array>
#include <cstdint>
#include <new>
#include <string>
#include <type_traits>
#include <utility>

namespace rust {
inline namespace cxxbridge05 {
// #include "rust/cxx.h"

#ifndef CXXBRIDGE05_PANIC
#define CXXBRIDGE05_PANIC
template <typename Exception>
void panic [[noreturn]] (const char *msg);
#endif // CXXBRIDGE05_PANIC

struct unsafe_bitcopy_t;

namespace {
template <typename T>
class impl;
} // namespace

#ifndef CXXBRIDGE05_RUST_STRING
#define CXXBRIDGE05_RUST_STRING
class String final {
public:
  String() noexcept;
  String(const String &) noexcept;
  String(String &&) noexcept;
  ~String() noexcept;

  String(const std::string &);
  String(const char *);
  String(const char *, size_t);

  String &operator=(const String &) noexcept;
  String &operator=(String &&) noexcept;

  explicit operator std::string() const;

  const char *data() const noexcept;
  size_t size() const noexcept;
  size_t length() const noexcept;

  String(unsafe_bitcopy_t, const String &) noexcept;

private:
  std::array<uintptr_t, 3> repr;
};
#endif // CXXBRIDGE05_RUST_STRING

#ifndef CXXBRIDGE05_RUST_STR
#define CXXBRIDGE05_RUST_STR
class Str final {
public:
  Str() noexcept;
  Str(const std::string &);
  Str(const char *);
  Str(const char *, size_t);
  Str(std::string &&) = delete;

  Str &operator=(const Str &) noexcept = default;

  explicit operator std::string() const;

  const char *data() const noexcept;
  size_t size() const noexcept;
  size_t length() const noexcept;

  Str(const Str &) noexcept = default;
  ~Str() noexcept = default;

private:
  friend impl<Str>;
  const char *ptr;
  size_t len;
};

inline const char *Str::data() const noexcept { return this->ptr; }

inline size_t Str::size() const noexcept { return this->len; }

inline size_t Str::length() const noexcept { return this->len; }
#endif // CXXBRIDGE05_RUST_STR

#ifndef CXXBRIDGE05_RUST_BITCOPY
#define CXXBRIDGE05_RUST_BITCOPY
struct unsafe_bitcopy_t final {
  explicit unsafe_bitcopy_t() = default;
};

constexpr unsafe_bitcopy_t unsafe_bitcopy{};
#endif // CXXBRIDGE05_RUST_BITCOPY

#ifndef CXXBRIDGE05_RUST_VEC
#define CXXBRIDGE05_RUST_VEC
template <typename T>
class Vec final {
public:
  using value_type = T;

  Vec() noexcept;
  Vec(Vec &&) noexcept;
  ~Vec() noexcept;

  Vec &operator=(Vec &&) noexcept;

  size_t size() const noexcept;
  bool empty() const noexcept;
  const T *data() const noexcept;
  T *data() noexcept;

  const T &operator[](size_t n) const noexcept;
  const T &at(size_t n) const;

  const T &front() const;
  const T &back() const;

  void reserve(size_t new_cap);
  void push_back(const T &value);
  void push_back(T &&value);
  template <typename... Args>
  void emplace_back(Args &&... args);

  class const_iterator final {
  public:
    using difference_type = ptrdiff_t;
    using value_type = typename std::add_const<T>::type;
    using pointer =
        typename std::add_pointer<typename std::add_const<T>::type>::type;
    using reference = typename std::add_lvalue_reference<
        typename std::add_const<T>::type>::type;
    using iterator_category = std::forward_iterator_tag;

    const T &operator*() const noexcept;
    const T *operator->() const noexcept;
    const_iterator &operator++() noexcept;
    const_iterator operator++(int) noexcept;
    bool operator==(const const_iterator &) const noexcept;
    bool operator!=(const const_iterator &) const noexcept;

  private:
    friend class Vec;
    const void *pos;
    size_t stride;
  };

  const_iterator begin() const noexcept;
  const_iterator end() const noexcept;

  Vec(unsafe_bitcopy_t, const Vec &) noexcept;

private:
  static size_t stride() noexcept;
  void reserve_total(size_t cap) noexcept;
  void set_len(size_t len) noexcept;
  void drop() noexcept;

  std::array<uintptr_t, 3> repr;
};

template <typename T>
Vec<T>::Vec(Vec &&other) noexcept {
  this->repr = other.repr;
  new (&other) Vec();
}

template <typename T>
Vec<T>::~Vec() noexcept {
  this->drop();
}

template <typename T>
Vec<T> &Vec<T>::operator=(Vec &&other) noexcept {
  if (this != &other) {
    this->drop();
    this->repr = other.repr;
    new (&other) Vec();
  }
  return *this;
}

template <typename T>
bool Vec<T>::empty() const noexcept {
  return size() == 0;
}

template <typename T>
T *Vec<T>::data() noexcept {
  return const_cast<T *>(const_cast<const Vec<T> *>(this)->data());
}

template <typename T>
const T &Vec<T>::operator[](size_t n) const noexcept {
  auto data = reinterpret_cast<const char *>(this->data());
  return *reinterpret_cast<const T *>(data + n * this->stride());
}

template <typename T>
const T &Vec<T>::at(size_t n) const {
  if (n >= this->size()) {
    panic<std::out_of_range>("rust::Vec index out of range");
  }
  return (*this)[n];
}

template <typename T>
const T &Vec<T>::front() const {
  return (*this)[0];
}

template <typename T>
const T &Vec<T>::back() const {
  return (*this)[this->size() - 1];
}

template <typename T>
void Vec<T>::reserve(size_t new_cap) {
  this->reserve_total(new_cap);
}

template <typename T>
void Vec<T>::push_back(const T &value) {
  this->emplace_back(value);
}

template <typename T>
void Vec<T>::push_back(T &&value) {
  this->emplace_back(std::move(value));
}

template <typename T>
template <typename... Args>
void Vec<T>::emplace_back(Args &&... args) {
  auto size = this->size();
  this->reserve_total(size + 1);
  ::new (reinterpret_cast<T *>(reinterpret_cast<char *>(this->data()) +
                               size * this->stride()))
      T(std::forward<Args>(args)...);
  this->set_len(size + 1);
}

template <typename T>
const T &Vec<T>::const_iterator::operator*() const noexcept {
  return *static_cast<const T *>(this->pos);
}

template <typename T>
const T *Vec<T>::const_iterator::operator->() const noexcept {
  return static_cast<const T *>(this->pos);
}

template <typename T>
typename Vec<T>::const_iterator &Vec<T>::const_iterator::operator++() noexcept {
  this->pos = static_cast<const uint8_t *>(this->pos) + this->stride;
  return *this;
}

template <typename T>
typename Vec<T>::const_iterator
Vec<T>::const_iterator::operator++(int) noexcept {
  auto ret = const_iterator(*this);
  this->pos = static_cast<const uint8_t *>(this->pos) + this->stride;
  return ret;
}

template <typename T>
bool Vec<T>::const_iterator::operator==(
    const const_iterator &other) const noexcept {
  return this->pos == other.pos;
}

template <typename T>
bool Vec<T>::const_iterator::operator!=(
    const const_iterator &other) const noexcept {
  return this->pos != other.pos;
}

template <typename T>
typename Vec<T>::const_iterator Vec<T>::begin() const noexcept {
  const_iterator it;
  it.pos = this->data();
  it.stride = this->stride();
  return it;
}

template <typename T>
typename Vec<T>::const_iterator Vec<T>::end() const noexcept {
  const_iterator it = this->begin();
  it.pos = static_cast<const uint8_t *>(it.pos) + it.stride * this->size();
  return it;
}

template <typename T>
Vec<T>::Vec(unsafe_bitcopy_t, const Vec &bits) noexcept : repr(bits.repr) {}
#endif // CXXBRIDGE05_RUST_VEC

template <typename T>
union MaybeUninit {
  T value;
  MaybeUninit() {}
  ~MaybeUninit() {}
};

namespace {
namespace repr {
struct PtrLen final {
  const void *ptr;
  size_t len;
};
} // namespace repr

template <>
class impl<Str> final {
public:
  static repr::PtrLen repr(Str str) noexcept {
    return repr::PtrLen{str.ptr, str.len};
  }
};
} // namespace
} // namespace cxxbridge05
} // namespace rust

namespace newsboat {
  namespace keymap {
    namespace bridged {
      struct Operation;
    }
  }
}

namespace newsboat {
namespace keymap {
namespace bridged {
extern "C" {
void newsboat$keymap$bridged$cxxbridge05$tokenize_operation_sequence(::rust::repr::PtrLen input, ::rust::Vec<::newsboat::keymap::bridged::Operation> *return$) noexcept;

const ::rust::Vec<::rust::String> *newsboat$keymap$bridged$cxxbridge05$operation_tokens(const ::newsboat::keymap::bridged::Operation &operation) noexcept;
} // extern "C"

::rust::Vec<::newsboat::keymap::bridged::Operation> tokenize_operation_sequence(::rust::Str input) noexcept {
  ::rust::MaybeUninit<::rust::Vec<::newsboat::keymap::bridged::Operation>> return$;
  newsboat$keymap$bridged$cxxbridge05$tokenize_operation_sequence(::rust::impl<::rust::Str>::repr(input), &return$.value);
  return ::std::move(return$.value);
}

const ::rust::Vec<::rust::String> &operation_tokens(const ::newsboat::keymap::bridged::Operation &operation) noexcept {
  return *newsboat$keymap$bridged$cxxbridge05$operation_tokens(operation);
}
} // namespace bridged
} // namespace keymap
} // namespace newsboat

extern "C" {
#ifndef CXXBRIDGE05_RUST_VEC_newsboat$keymap$bridged$Operation
#define CXXBRIDGE05_RUST_VEC_newsboat$keymap$bridged$Operation
void cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$new(const ::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr) noexcept;
void cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$drop(::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr) noexcept;
size_t cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$len(const ::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr) noexcept;
const ::newsboat::keymap::bridged::Operation *cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$data(const ::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr) noexcept;
void cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$reserve_total(::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr, size_t cap) noexcept;
void cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$set_len(::rust::Vec<::newsboat::keymap::bridged::Operation> *ptr, size_t len) noexcept;
size_t cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$stride() noexcept;
#endif // CXXBRIDGE05_RUST_VEC_newsboat$keymap$bridged$Operation
} // extern "C"

namespace rust {
inline namespace cxxbridge05 {
template <>
Vec<::newsboat::keymap::bridged::Operation>::Vec() noexcept {
  cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$new(this);
}
template <>
void Vec<::newsboat::keymap::bridged::Operation>::drop() noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$drop(this);
}
template <>
size_t Vec<::newsboat::keymap::bridged::Operation>::size() const noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$len(this);
}
template <>
const ::newsboat::keymap::bridged::Operation *Vec<::newsboat::keymap::bridged::Operation>::data() const noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$data(this);
}
template <>
void Vec<::newsboat::keymap::bridged::Operation>::reserve_total(size_t cap) noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$reserve_total(this, cap);
}
template <>
void Vec<::newsboat::keymap::bridged::Operation>::set_len(size_t len) noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$set_len(this, len);
}
template <>
size_t Vec<::newsboat::keymap::bridged::Operation>::stride() noexcept {
  return cxxbridge05$rust_vec$newsboat$keymap$bridged$Operation$stride();
}
} // namespace cxxbridge05
} // namespace rust
