#include <array>
#include <cstdint>
#include <new>
#include <string>
#include <type_traits>
#include <utility>

namespace rust {
inline namespace cxxbridge05 {
// #include "rust/cxx.h"

struct unsafe_bitcopy_t;

namespace {
template <typename T>
class impl;
} // namespace

#ifndef CXXBRIDGE05_RUST_STRING
#define CXXBRIDGE05_RUST_STRING
class String final {
public:
  String() noexcept;
  String(const String &) noexcept;
  String(String &&) noexcept;
  ~String() noexcept;

  String(const std::string &);
  String(const char *);
  String(const char *, size_t);

  String &operator=(const String &) noexcept;
  String &operator=(String &&) noexcept;

  explicit operator std::string() const;

  const char *data() const noexcept;
  size_t size() const noexcept;
  size_t length() const noexcept;

  String(unsafe_bitcopy_t, const String &) noexcept;

private:
  std::array<uintptr_t, 3> repr;
};
#endif // CXXBRIDGE05_RUST_STRING

#ifndef CXXBRIDGE05_RUST_STR
#define CXXBRIDGE05_RUST_STR
class Str final {
public:
  Str() noexcept;
  Str(const std::string &);
  Str(const char *);
  Str(const char *, size_t);
  Str(std::string &&) = delete;

  Str &operator=(const Str &) noexcept = default;

  explicit operator std::string() const;

  const char *data() const noexcept;
  size_t size() const noexcept;
  size_t length() const noexcept;

  Str(const Str &) noexcept = default;
  ~Str() noexcept = default;

private:
  friend impl<Str>;
  const char *ptr;
  size_t len;
};

inline const char *Str::data() const noexcept { return this->ptr; }

inline size_t Str::size() const noexcept { return this->len; }

inline size_t Str::length() const noexcept { return this->len; }
#endif // CXXBRIDGE05_RUST_STR

#ifndef CXXBRIDGE05_RUST_BOX
#define CXXBRIDGE05_RUST_BOX
template <typename T>
class Box final {
public:
  using value_type = T;
  using const_pointer =
      typename std::add_pointer<typename std::add_const<T>::type>::type;
  using pointer = typename std::add_pointer<T>::type;

  Box(const Box &);
  Box(Box &&) noexcept;
  ~Box() noexcept;

  explicit Box(const T &);
  explicit Box(T &&);

  Box &operator=(const Box &);
  Box &operator=(Box &&) noexcept;

  const T *operator->() const noexcept;
  const T &operator*() const noexcept;
  T *operator->() noexcept;
  T &operator*() noexcept;

  template <typename... Fields>
  static Box in_place(Fields &&...);

  static Box from_raw(T *) noexcept;

  T *into_raw() noexcept;

private:
  Box() noexcept;
  void uninit() noexcept;
  void drop() noexcept;
  T *ptr;
};

template <typename T>
Box<T>::Box(const Box &other) : Box(*other) {}

template <typename T>
Box<T>::Box(Box &&other) noexcept : ptr(other.ptr) {
  other.ptr = nullptr;
}

template <typename T>
Box<T>::Box(const T &val) {
  this->uninit();
  ::new (this->ptr) T(val);
}

template <typename T>
Box<T>::Box(T &&val) {
  this->uninit();
  ::new (this->ptr) T(std::move(val));
}

template <typename T>
Box<T>::~Box() noexcept {
  if (this->ptr) {
    this->drop();
  }
}

template <typename T>
Box<T> &Box<T>::operator=(const Box &other) {
  if (this != &other) {
    if (this->ptr) {
      **this = *other;
    } else {
      this->uninit();
      ::new (this->ptr) T(*other);
    }
  }
  return *this;
}

template <typename T>
Box<T> &Box<T>::operator=(Box &&other) noexcept {
  if (this->ptr) {
    this->drop();
  }
  this->ptr = other.ptr;
  other.ptr = nullptr;
  return *this;
}

template <typename T>
const T *Box<T>::operator->() const noexcept {
  return this->ptr;
}

template <typename T>
const T &Box<T>::operator*() const noexcept {
  return *this->ptr;
}

template <typename T>
T *Box<T>::operator->() noexcept {
  return this->ptr;
}

template <typename T>
T &Box<T>::operator*() noexcept {
  return *this->ptr;
}

template <typename T>
template <typename... Fields>
Box<T> Box<T>::in_place(Fields &&... fields) {
  Box box;
  box.uninit();
  ::new (box.ptr) T{std::forward<Fields>(fields)...};
  return box;
}

template <typename T>
Box<T> Box<T>::from_raw(T *raw) noexcept {
  Box box;
  box.ptr = raw;
  return box;
}

template <typename T>
T *Box<T>::into_raw() noexcept {
  T *raw = this->ptr;
  this->ptr = nullptr;
  return raw;
}

template <typename T>
Box<T>::Box() noexcept {}
#endif // CXXBRIDGE05_RUST_BOX

namespace {
namespace repr {
struct PtrLen final {
  const void *ptr;
  size_t len;
};
} // namespace repr

template <>
class impl<Str> final {
public:
  static repr::PtrLen repr(Str str) noexcept {
    return repr::PtrLen{str.ptr, str.len};
  }
};
} // namespace
} // namespace cxxbridge05
} // namespace rust

namespace newsboat {
  namespace scopemeasure {
    namespace bridged {
      struct ScopeMeasure;
    }
  }
}

namespace newsboat {
namespace scopemeasure {
namespace bridged {
#ifndef CXXBRIDGE05_STRUCT_newsboat$scopemeasure$bridged$ScopeMeasure
#define CXXBRIDGE05_STRUCT_newsboat$scopemeasure$bridged$ScopeMeasure
struct ScopeMeasure final {
  ScopeMeasure() = delete;
  ScopeMeasure(const ScopeMeasure &) = delete;
  void stopover(::rust::Str stopover_name) const noexcept;
};
#endif // CXXBRIDGE05_STRUCT_newsboat$scopemeasure$bridged$ScopeMeasure

extern "C" {
::newsboat::scopemeasure::bridged::ScopeMeasure *newsboat$scopemeasure$bridged$cxxbridge05$create(::rust::String *scope_name) noexcept;

void newsboat$scopemeasure$bridged$cxxbridge05$ScopeMeasure$stopover(const ::newsboat::scopemeasure::bridged::ScopeMeasure &self, ::rust::repr::PtrLen stopover_name) noexcept;
} // extern "C"

::rust::Box<::newsboat::scopemeasure::bridged::ScopeMeasure> create(::rust::String scope_name) noexcept {
  return ::rust::Box<::newsboat::scopemeasure::bridged::ScopeMeasure>::from_raw(newsboat$scopemeasure$bridged$cxxbridge05$create(&scope_name));
}

void ScopeMeasure::stopover(::rust::Str stopover_name) const noexcept {
  newsboat$scopemeasure$bridged$cxxbridge05$ScopeMeasure$stopover(*this, ::rust::impl<::rust::Str>::repr(stopover_name));
}
} // namespace bridged
} // namespace scopemeasure
} // namespace newsboat

extern "C" {
#ifndef CXXBRIDGE05_RUST_BOX_newsboat$scopemeasure$bridged$ScopeMeasure
#define CXXBRIDGE05_RUST_BOX_newsboat$scopemeasure$bridged$ScopeMeasure
void cxxbridge05$box$newsboat$scopemeasure$bridged$ScopeMeasure$uninit(::rust::Box<::newsboat::scopemeasure::bridged::ScopeMeasure> *ptr) noexcept;
void cxxbridge05$box$newsboat$scopemeasure$bridged$ScopeMeasure$drop(::rust::Box<::newsboat::scopemeasure::bridged::ScopeMeasure> *ptr) noexcept;
#endif // CXXBRIDGE05_RUST_BOX_newsboat$scopemeasure$bridged$ScopeMeasure
} // extern "C"

namespace rust {
inline namespace cxxbridge05 {
template <>
void Box<::newsboat::scopemeasure::bridged::ScopeMeasure>::uninit() noexcept {
  cxxbridge05$box$newsboat$scopemeasure$bridged$ScopeMeasure$uninit(this);
}
template <>
void Box<::newsboat::scopemeasure::bridged::ScopeMeasure>::drop() noexcept {
  cxxbridge05$box$newsboat$scopemeasure$bridged$ScopeMeasure$drop(this);
}
} // namespace cxxbridge05
} // namespace rust
